#
# READ or WRITE POSSIBLY ENCRYPTED PYTHON CODE FILES
#

from __future__ import print_function
from shlib import Path, mkdir
from messenger import display, error, narrate, debug, os_error
from .preferences import (
    SETTINGS_DIR, DEFAULT_ACCOUNTS_FILENAME, DEFAULT_LOG_FILENAME, 
    DEFAULT_ARCHIVE_FILENAME
)

class AccountFile:
    def __init__(self, path, gpg, generator, init=None, contents=''):
        path = Path(path).expanduser()
        try:
            mkdir(Path(SETTINGS_DIR).expanduser())
            if init and path.exists():
                display("%s: already exists." % path)
                # file creation (init) requested, but file already exists
                # don't overwrite the file, instead read it so the information 
                # can be used to create any remaining files.
            if init and not path.exists():
                # create the file
                code = contents.format(
                    dict_hash='not implemented yet',
                    secrets_hash='not implemented yet',
                    charsets_hash='not implemented yet',
                    accounts_file=DEFAULT_ACCOUNTS_FILENAME,
                    log_file=DEFAULT_LOG_FILENAME,
                    archive_file=DEFAULT_ARCHIVE_FILENAME,
                    gpg_id=gpg.gpg_id,
                    gpg_home='~/.gnupg',
                    gpg_path='/usr/bin/gpg2',
                    section='{''{''{''1',
                    master_password='not implemented yet',
                )
                display('%s: creating.' % path)
                if path.suffix in ['.gpg', '.asc']:
                    narrate('encrypting.', culprit=path)
                    # encrypt it
                    gpg.save(Path(path), code)
                else:
                    narrate('not encrypting.', culprit=path)
                    # file is not encrypted
                    with path.open('w') as f:
                        f.write(code)
            else:
                # read the file
                if path.suffix in ['.gpg', '.asc']:
                    # file is encrypted, decrypt it
                    code = gpg.read(Path(path))
                else:
                    # file is not encrypted
                    code = path.read_text()
        except OSError as exception:
            error(os_error(exception))

        contents = {}
        compiled = compile(code, str(path), 'exec')
        exec(compiled, contents)
        if 'master_password' in contents:
            generator.add_missing_master(contents['master_password'])
        self.contents = contents

    def __getattr__(self, name):
        return self.contents[name]



from __future__ import print_function
from messenger import log
from .preferences import SEARCH_FIELDS

class Account(type):
    def __init__(self):
        raise NotImplementedError

    @classmethod
    def all_accounts(cls):
        for sub in cls.__subclasses__(cls):
            yield sub
            for sub in sub.all_accounts():
                yield sub

    @classmethod
    def get_name(cls):
        try:
            return cls.name
        except AttributeError:
            # consider converting lower to upper case transitions in __name__ to 
            # dashes.
            return cls.__name__.lower()

    @classmethod
    def matches_exactly(cls, account):
        if account == cls.get_name():
            return True
        try:
            if account in cls.aliases:
                return True
        except AttributeError:
            pass
        return False

    @classmethod
    def id_contains(cls, target):
        if target in cls.get_name():
            return True
        try:
            for alias in cls.aliases:
                if target in alias:
                    return True
        except AttributeError:
            pass
        return False

    @classmethod
    def account_contains(cls, target):
        if cls.id_contains(target):
            return True
        for field in SEARCH_FIELDS:
            try:
                if target in cls.__dict__[field]:
                    return True
            except KeyError:
                pass
        return False

    @classmethod
    def initialize(cls):
        log('initializing', cls.get_name())
        for key, value in cls.__dict__.items():
            if not key.startswith('__'):
                log('    initializing attribute', key)
                try:
                    value._initiate(key, cls)
                except AttributeError:
                    try:
                        for i, each in enumerate(value):
                            name = 'key[%s]' % i
                            each._initiate(name, cls)
                    except AttributeError:
                        pass

    @classmethod
    def values(cls):
        for key in sorted(cls.__dict__):
            if not key.startswith('__'):
                yield key, cls.__dict__[key]

    @classmethod
    def get_value(cls, name):
        return cls.__dict__.get(name)



# Avendesora Password Generator Preferences
#
# Copyright (C) 2016 Kenneth S. Kundert

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

# Imports {{{1
from textwrap import dedent
import re

# Filenames {{{1
SETTINGS_DIR = '~/.config/avendesora'
DICTIONARY_FILENAME = 'words'
CONFIG_FILENAME = 'config'
    # config file must be unencrypted (it contains the gpg settings)
DEFAULT_ACCOUNTS_FILENAME = 'accounts.gpg'
DEFAULT_LOG_FILENAME = 'log.gpg'
    # log file will be encrypted if you add .gpg or .asc extension
DEFAULT_ARCHIVE_FILENAME = 'archive.gpg'

# Settings {{{1
LABEL_COLOR = 'yellow'
    # choose from normal, black, red, green, yellow, blue, magenta, cyan, white
LABEL_STYLE = 'normal'
    # choose from normal, bright, reverse, dim, underline, blink, reverse,
    # invisible (these need to be implemented by underlying terminal, and some
    # are not (such a blink and dim)
INITIAL_AUTOTYPE_DELAY = 0.0
DEBUG = False
    # Turns on the logging of extra information, but may expose sensitive
    # account information in the log file.
PREFER_HTTPS = True
    # When PREFER_HTTPS is true, avendesora requires the https protocol unless
    # http is explicitly specified in the url.
    # When PREFER_HTTPS is false, avendesora allows the http protocol unless
    # https is explicitly specified in the url.


# Utility programs {{{1
# use absolute paths for xdotool and xsel so they cannot be maliciously replaced
XDOTOOL = '/usr/bin/xdotool'
XSEL = '/usr/bin/xsel'
NOTIFIER_NORMAL = ['notify-send', '--urgency=low']
NOTIFIER_ERROR = ['notify-send', '--urgency=normal']

# Signatures {{{1
# These signatures must be the sha1 signatures for the corresponding files
# Regenerate them with 'sha1sum <filename>'
# These are used in creating the initial master password file.
SECRETS_SHA1 = "5d1c97a0fb699241fca5d50a7ad0508047990510"
CHARSETS_SHA1 = "dab48b2103ebde97f78cfebd15cc1e66d6af6ed0"
DICTIONARY_SHA1 = "d9aa1c08e08d6cacdf82819eeb5832429eadb95a"

# GPG Settings
GPG_PATH = '/usr/bin/gpg2'
GPG_HOME = '~/.gnupg'
GPG_ARMOR = True

# Browsers {{{1
# Associate a command with a browser key.
# The command must contain a single %s, which is replaced with URL.
BROWSERS = {
    'x': 'xdg-open %s > /dev/null', # system default browser
    'f': 'firefox -new-tab %s > /dev/null',
    'd': 'dwb %s',
    'c': 'google-chrome %s',
    't': 'torbrowser %s > /dev/null',
}
DEFAULT_BROWSER = 'x'


# Account Recognition {{{1
# Title Recognition
# Build up the regular expression used to recognize the various component of 
# the window title.
def labelRegex(label, regex):
    return "(?P<%s>%s)" % (label, regex)
HOST_REGEX = r'(?:[a-zA-Z0-9\-]+\.)+[a-zA-Z0-9]+'
EMAIL_REGEX = r'[a-zA-Z0-9\-]+@' + HOST_REGEX
REGEX_COMPONENTS = {
    # If you add new components, you must also add code that handles the
    # component in accounts.py.
    'title': labelRegex('title', r'.*'),
    'host': labelRegex('host', HOST_REGEX),
    'protocol': labelRegex('protocol', r'\w+'),
    'browser': labelRegex('browser', r'\w+'),
    'username': labelRegex('username', r'\w+'),
    'email': labelRegex('email', EMAIL_REGEX)}
# Hostname in Titlebar browser title regex
HNITB_BROWSER_TITLE_PATTERN = re.compile(
    r'(?:{title} - )?{host} \({protocol}\)(?: - {browser})?'.format(
        **REGEX_COMPONENTS
    )
)
# This is for version 3 and beyond; requires that preferences in HNINTB be set 
# to 'show the short URL' with a separator of '-'.
HNITBv3_BROWSER_TITLE_PATTERN = re.compile(
    r'(?:{title} - ){protocol}?://{host}(?: - {browser})?'.format(
        **REGEX_COMPONENTS
    )
)
# Simple browser title regex
SIMPLE_BROWSER_TITLE_PATTERN = re.compile(
    r'{title}(?: - {browser})?'.format(**REGEX_COMPONENTS))
# Recognize components of the url
URL_PATTERN = re.compile(
    r'(?:{protocol}://)?{host}(?:/.*)?'.format(**REGEX_COMPONENTS))
TITLE_PATTERNS = [
    ('hostname-in-titlebar-browser-v3', HNITBv3_BROWSER_TITLE_PATTERN),
    #('hostname-in-titlebar-browser', HNITB_BROWSER_TITLE_PATTERN),
    # You can comment out the entry above if you are not using 'Hostname in
    # Titlebar' extension to Firefox and Thunderbird
    ('simple browser title', SIMPLE_BROWSER_TITLE_PATTERN)]

# Initial config file {{{1
CONFIG_FILE_INITIAL_CONTENTS = dedent('''\
    # Changing the contents of the dictionary, secrets, or charsets will change 
    # the secrets you generate, thus you do not want to change these files once 
    # you have started using the program. These are hashes for the contents of 
    # these files at the time that this file was created. The program will 
    # complain if the current contents of these files generate a different 
    # hash. Only update these hashes if you know what you are doing.
    dict_hash = '{dict_hash}'      # DO NOT CHANGE THIS LINE
    secrets_hash = '{secrets_hash}'   # DO NOT CHANGE THIS LINE
    charsets_hash = '{charsets_hash}'  # DO NOT CHANGE THIS LINE

    # List of the files that contain account information
    accounts_files = ['{accounts_file}']

    # The desired location of the log file (relative to config directory).
    # Adding a suffix of .gpg or .asc causes the file to be encrypted 
    # (otherwise it can leak account names). Use None to disable logging.
    log_file = '{log_file}'

    # The desired location of the archive file (relative to config directory 
    # and end the path in .gpg). Use None to disable archiving.
    archive_file = '{archive_file}'

    # Information used by GPG when encrypting and decrypting files.
    gpg_id = '{gpg_id}'
    gpg_home = '{gpg_home}'
    gpg_path = '{gpg_path}'

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')


# Initial accounts file {{{1
ACCOUNTS_FILE_INITIAL_CONTENTS = dedent('''\
    # Account information
    #
    # Add information about each of your accounts to the accounts dictionary.
    #
    # You can use the dedent function to strip leading whitespace from
    # multi-line remarks.  You can use the character sets and exclude function
    # to create alphabets for you character-base passwords.
    #
    # Example:
    # To create an alphabet with all characters except tabs use either:
    #     'alphabet': exclude(PRINTABLE, '\\t')
    # or:
    #     'alphabet': ALPHANUMERIC + PUNCTUATION + ' '

    from textwrap import dedent
    from avendesora.charsets import (
        exclude, LOWERCASE, UPPERCASE, LETTERS, DIGITS, ALPHANUMERIC,
        HEXDIGITS, PUNCTUATION, WHITESPACE, PRINTABLE, DISTINGUISHABLE
    )
    from avendesora.utilities import gethostname, getusername
    from avendesora.account import Account
    from avendesora.secrets import Secret, Hidden

    master_password = Hidden("""{master_password}""")

    # Accounts
    # AAA {section}
    # BBB {section}
    # CCC {section}
    # DDD {section}
    # EEE {section}
    # FFF {section}
    # GGG {section}
    # HHH {section}
    # III {section}
    # JJJ {section}
    # KKK {section}
    # LLL {section}
    # MMM {section}
    # NNN {section}
    # OOO {section}
    # PPP {section}
    # QQQ {section}
    # RRR {section}
    # SSS {section}
    # TTT {section}
    # UUU {section}
    # VVV {section}
    # WWW {section}
    # XXX {section}
    # YYY {section}
    # ZZZ {section}

    # vim: filetype=python sw=4 sts=4 et ai ff=unix :
''')

# Fields {{{1
# Do not change these (not user configurable)
SEARCH_FIELDS = ['username', 'account', 'email', 'url', 'remarks']
STRING_FIELDS = [
    'alphabet', 'autotype', 'email', 'master', 'prefix',
    'remarks', 'separator', 'suffix', 'template', 'type',
    'username', 'version'
]
INTEGER_FIELDS = ['num-chars', 'num-words']
LIST_FIELDS = ['security questions', 'aliases']
LIST_OR_STRING_FIELDS = ['account', 'window', 'url']
ENUM_FIELDS = {
    'password-type': ['words', 'chars']
}
ALL_FIELDS = (
    STRING_FIELDS + INTEGER_FIELDS + LIST_FIELDS + LIST_OR_STRING_FIELDS +
    [each for each in ENUM_FIELDS.keys()]
)



#
# INTERFACE TO GNUPG PACKAGE
#

from .preferences import GPG_PATH, GPG_HOME, GPG_ARMOR
from messenger import display, error, fatal, is_collection
from shlib import Path
import gnupg
import io


class GPG:
    def __init__(self,
        gpg_id=None, gpg_path=None, gpg_home=None, armor=None
    ):
        self.gpg_id = gpg_id if gpg_id else self._guess_id()
        self.gpg_path = Path(gpg_path if gpg_path else GPG_PATH)
        self.gpg_home = Path(gpg_home if gpg_path else GPG_HOME).expanduser()
        self.armor = armor if armor is not None else GPG_ARMOR

        gpg_args = {}
        if self.gpg_path:
            gpg_args.update({'gpgbinary': str(self.gpg_path)})
        if self.gpg_home:
            gpg_args.update({'gnupghome': str(self.gpg_home)})
        self.gpg = gnupg.GPG(**gpg_args)

    def update_id(self, gpg_id):
        self.gpg_id = gpg_id

    def _guess_id(self):
        import socket, getpass
        username = getpass.getuser()
        hostname = socket.gethostname().split('.')
        if len(hostname) <= 2:
            hostname = '.'.join(hostname)
        else:
            # strip off name of local machine
            hostname = '.'.join(hostname[1:])
        return username + '@' + hostname

    def save(self, path, contents):
        encrypted = self.gpg.encrypt(contents, self.gpg_id, armor=self.armor)
        if not encrypted.ok:
            fatal('unable to encrypt.', encrypted.stderr, culprit=path, sep='\n')
        else:
            if self.armor:
                path.write_text(str(encrypted))
            else:
                path.write_bytes(encrypted)
            path.chmod(0o600)

    def read(self, path):
        with path.open('rb') as f:
            decrypted = self.gpg.decrypt_file(f)
            if not decrypted.ok:
                fatal('unable to decrypt.', decrypted.stderr, culprit=path, sep='\n')
        return decrypted.data

    def open(self, path):
        self.path = path
        self.stream = io.StringIO()
        return self.stream

    def close(self):
        contents = self.stream.getvalue()
        self.save(self.path, contents)



#!/usr/bin/env python3
"""
Avendesora Password Generator

Generates passwords and pass phrases based on stored account information.

usage:
    avendesora [options]
    avendesora [options] <account> [<secret>]

options:
    -f <text>, --find <text>
                            List any account that contains the given string in 
                            its ID or aliases.
    -g <GPG ID>, --gpgid <GPG ID>
                            Use this GPG ID when creating any missing files.
    -H <plain_text>, --hide <plain_text>
                            Encode plain_text using base64 as a way of hiding 
                            (but not encrypting) its value.
    -R <coded_text>, --reveal <coded_text>
                            Decode coded_text using base64.
    -s <text>, --search <text>
                            List any account that contains the given string in 
                            {search_fields} or its ID.
"""

# License {{{1
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .generator import PasswordGenerator
from .gpg import GPG
from .preferences import SETTINGS_DIR, DEFAULT_LOG_FILENAME, SEARCH_FIELDS
from .utilities import Hidden
from messenger import Messenger, Error, output, terminate, debug
from shlib import Path

import docopt
import sys

# Utilities {{{1
def print_search_results(search_term, search_func):
    to_print = []
    for acct in search_func(search_term):
        aliases = getattr(acct, 'aliases', [])

        aliases = ' (%s)' % (', '.join(aliases)) if aliases else ''
        to_print += [acct.get_name() + aliases]
    output(search_term + ':')
    output('    ' + ('\n    '.join(sorted(to_print))))

# Main {{{1
def main():
    cmdline = docopt.docopt(
        __doc__.format(search_fields=', '.join(SEARCH_FIELDS))
    )
    if cmdline['--hide']:
        debug(cmdline['--hide'])
        output(Hidden.hide(cmdline['--hide']))
        terminate()
    if cmdline['--reveal']:
        output(Hidden.reveal(cmdline['--reveal']))
        terminate()

    def teardown():
        messenger.disconnect()
        gpg.close()

    gpg = GPG()
    messenger = Messenger(
        logfile=gpg.open(Path(SETTINGS_DIR, DEFAULT_LOG_FILENAME).expanduser()),
        termination_callback=teardown
    )
    try:
        generator = PasswordGenerator(
            gpg_id = cmdline['--gpgid']
        )
        if cmdline['--find']:
            print_search_results(cmdline['--find'], generator.find_accounts)
            terminate()

        if cmdline['--search']:
            print_search_results(cmdline['--search'], generator.search_accounts)
            terminate()

        generator.activate_account(cmdline['<account>'])
        secret = generator.get_secret(cmdline['<secret>'])
        print('%s = %s' % (cmdline['<secret>'], secret))
    except KeyboardInterrupt:
        output('Terminated by user.')
    except Error as err:
        err.terminate()

    terminate()



from .preferences import (
    SETTINGS_DIR, CONFIG_FILENAME, CONFIG_FILE_INITIAL_CONTENTS,
    ACCOUNTS_FILE_INITIAL_CONTENTS
)
from .files import AccountFile
from .gpg import GPG
from .dictionary import DICTIONARY
from .account import Account
from messenger import Error, debug, terminate_if_errors
from shlib import Path

class PasswordGenerator:
    def __init__(self, init=False, gpg_id=None):
        self.gpg = GPG(gpg_id)
        self.active_account = None

        # First open the config file
        self.config = AccountFile(
            Path(SETTINGS_DIR, CONFIG_FILENAME),
            self.gpg,
            self,
            init,
            CONFIG_FILE_INITIAL_CONTENTS,
        )

        # Now open any accounts files found
        for filename in self.config.accounts_files:
            AccountFile(
                Path(SETTINGS_DIR, filename),
                self.gpg,
                self,
                init,
                ACCOUNTS_FILE_INITIAL_CONTENTS,
            )
        terminate_if_errors()

        DICTIONARY.validate(self.config.dict_hash)

    def activate_account(self, account):
        if not account:
            raise Error('no account specified.')
        for each in Account.all_accounts():
            if each.matches_exactly(account):
                each.initialize()
                self.active_account = each
                return
        raise Error('not found', culprit=account)

    def get_secret(self, secret=None):
        if not secret:
             secret = self.active_account.get_value('default')
        if not secret:
             secret = 'passcode'
        value = self.active_account.get_value(secret)
        if value:
            return value
        else:
            raise Error('not found', culprit=secret)

    def find_accounts(self, target):
        accounts = []
        for each in Account.all_accounts():
            if each.id_contains(target):
                accounts.append(each)
        return accounts

    def search_accounts(self, target):
        accounts = []
        for each in Account.all_accounts():
            if each.account_contains(target):
                accounts.append(each)
        return accounts

    def add_missing_master(self, master):
        for each in Account.all_accounts():
            if not hasattr(each, 'master'):
                each.master = master



from __future__ import print_function
from binascii import a2b_base64, b2a_base64, Error
from messenger import error, is_str, log

# gethostname {{{1
# returns short version of the hostname (the hostname without any domain name)
import socket
def gethostname():
    return socket.gethostname().split('.')[0]

# getusername {{{1
import getpass
def getusername():
    return getpass.getuser()

# RecognizeURL {{{1
class RecognizeURL():
    def __init__(self, *args, **kwargs):
        pass
    def _initiate(self, name, account):
        log('        initializing RecognizeURL(%s)' % account.get_name())

# Autotype {{{1
class Autotype():
    def __init__(self, *args, **kwargs):
        pass
    def _initiate(self, name, account):
        log('        initializing Autotype(%s)' % account.get_name())

# Hidden {{{1
import linecache
import sys

class Hidden():
    def __init__(self, value, **kwargs):
        try:
            value = a2b_base64(value)
            self.value = value.decode(kwargs.get('encoding', 'utf8'))
        except Error as err:
            import traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            filename, lineno = traceback.extract_stack()[-2][:2]
                # context and content are also available, but in this case
                # Hidden is generally instantiated from top-level so the 
                # context is not interesting and the content (the actual line 
                # of code) shown in this case is gibberish (encrypted).
            error(
                'invalid value specified to Hidden().',
                culprit=(filename, lineno)
            )

    def _initiate(self, name, account):
        log('        initializing Hidden(%s)' % account.get_name())

    def __str__(self):
        return self.value

    @staticmethod
    def hide(value, **kwargs):
        value = value.encode(kwargs.get('encoding', 'utf8'))
        return b2a_base64(value).rstrip().decode('ascii')

    @staticmethod
    def reveal(value, **kwargs):
        value = a2b_base64(value.encode('ascii'))
        return value.decode(kwargs.get('encoding', 'utf8'))




#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.rst') as file:
    readme = file.read()

setup(
    name='avendesora',
    version='0.0.2',
    author='Ken Kundert and Kale Kundert',
    author_email='avendesora@nurdletech.com',
    description='An XKCD-style password generator.',
    long_description=readme,
    url='https://github.com/kenkundert/avendesora',
    include_package_data=True,
    license='GPLv3+',
    zip_safe=False,
    keywords=[
        'avendesora',
        'password',
        'XKCD',
    ],
    packages=[
        'avendesora',
    ],
    data_files=[('avendesora', ['avendesora/words'])],
    install_requires=[
        'docopt',
        'python-gnupg',
        'messenger',
        'shlib',
    ],
    entry_points = {
        'console_scripts': ['avendesora=avendesora.cli:main'],
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.5',
        'Topic :: Security :: Cryptography',
        'Topic :: Utilities',
    ],
)
